from __future__ import annotations

"""Helpers for reading unpacked Darwin Core Archives (DwC-A)."""

from pathlib import Path
from typing import Any, List
import xml.etree.ElementTree as ET

import polars as pl

from .lf_csv import DarwinCoreCsvLazyFrame

__all__ = ["scan_archive"]


def _parse_meta(meta_path: Path) -> tuple[str, bool, str, List[str]]:
    """Return information (core_file, has_header, separator, column_names)."""

    tree = ET.parse(meta_path)
    root = tree.getroot()

    # Handle XML namespace if present
    ns = {"dwc": "http://rs.tdwg.org/dwc/text/"}

    # Try with namespace first, then without
    core_elem = root.find("dwc:core", ns)
    if core_elem is None:
        core_elem = root.find(".//core")
    if core_elem is None:
        raise ValueError("meta.xml does not contain <core> element")

    # file location – in <files><location>relative/path</location></files>
    files_elem = core_elem.find(".//files")
    if files_elem is None:
        files_elem = core_elem.find("dwc:files", ns)
    if files_elem is None:
        raise ValueError("<core> missing <files>")

    location_elem = files_elem.find(".//location")
    if location_elem is None:
        location_elem = files_elem.find("dwc:location", ns)
    if location_elem is None or not location_elem.text:
        raise ValueError("<files> missing <location>")
    core_file = location_elem.text.strip()

    # delimiter & header
    separator = core_elem.get("fieldsTerminatedBy", "\t")
    # XML may encode tab as "\t" literal or as actual tab char
    if separator == "\t":
        separator = "\t"
    elif separator == "\\t":
        separator = "\t"

    ignore_header = int(core_elem.get("ignoreHeaderLines", "0"))
    has_header = ignore_header >= 1

    # column order
    fields: List[str] = []
    field_elems = core_elem.findall(".//field")
    if not field_elems:
        field_elems = core_elem.findall("dwc:field", ns)

    for field_elem in field_elems:
        index_str = field_elem.get("index")
        term_uri = field_elem.get("term")
        if index_str is None or term_uri is None:
            continue
        try:
            idx = int(index_str)
        except ValueError:
            continue
        # extract local term name from URI
        term = term_uri.rsplit("/", 1)[-1].rsplit("#", 1)[-1]
        if len(fields) <= idx:
            fields.extend([""] * (idx - len(fields) + 1))
        fields[idx] = term

    # some meta.xml include <id index="0" /> that represents the record id
    id_elem = core_elem.find(".//id")
    if id_elem is None:
        id_elem = core_elem.find("dwc:id", ns)
    assert id_elem is not None
    idx2 = id_elem.get("index")
    if idx2 is not None:
        idx = int(idx2)
        if len(fields) <= idx:
            fields.extend([""] * (idx - len(fields) + 1))
        # id doesn't have a term; choose "id"
        if not fields[idx]:
            fields[idx] = "id"

    # fill any empty column names with fallback names
    fields = [name if name else f"col_{i}" for i, name in enumerate(fields)]

    return core_file, has_header, separator, fields


def scan_archive(path: str | Path, **scan_csv_kwargs: Any) -> DarwinCoreCsvLazyFrame:  # noqa: D401
    """Scan an *unpacked* Darwin Core Archive directory lazily.

    Parameters
    ----------
    path:
        Path to a directory that contains at least ``meta.xml`` and the core
        data file referenced from it.
    **scan_csv_kwargs:
        Extra keyword arguments forwarded to :pyfunc:`polars.scan_csv` (e.g.
        ``infer_schema_length``).

    Returns
    -------
    DarwinCoreCsvLazyFrame
    """

    base_dir = Path(path)
    meta_path = base_dir / "meta.xml"
    if not meta_path.exists():
        raise FileNotFoundError("meta.xml not found in archive directory")

    core_file_rel, has_header, separator, columns = _parse_meta(meta_path)
    data_path = base_dir / core_file_rel

    inner = pl.scan_csv(
        data_path,
        separator=separator,
        has_header=has_header,
        new_columns=columns if has_header is False else None,
        **scan_csv_kwargs,
    )

    return DarwinCoreCsvLazyFrame(inner)


from __future__ import annotations

from pathlib import Path
from typing import Any, Mapping, Sequence

import polars as pl

__all__ = ["DarwinCoreDataFrame", "read_darwin_core_dataframe"]


def read_darwin_core_dataframe(path: str | Path, **read_csv_kwargs: Any) -> "DarwinCoreDataFrame":
    """Read a Darwin Core CSV eagerly into a *polars* DataFrame.

    Parameters
    ----------
    path:
        File path to a Darwin Core‐compatible CSV file.
    **read_csv_kwargs:
        Additional keyword arguments forwarded to :pyfunc:`polars.read_csv`.

    Returns
    -------
    DarwinCoreDataFrame
    """

    return DarwinCoreDataFrame(pl.read_csv(path, **read_csv_kwargs))


class DarwinCoreDataFrame(pl.DataFrame):
    """A *polars* DataFrame specialised for Darwin Core data.

    This subclass does not currently add new behaviour; it exists mainly for
    type clarity and to provide a future-proof extension point.
    """

    # NOTE: Sub-classing ``polars.DataFrame`` is *not* officially supported, but
    # in practice works fine for simple wrapper use-cases.  We keep the class
    # intentionally thin to avoid surprises.

    # Make DataFrame constructor inherits base.

    # You can extend with domain-specific helpers later, for example:
    #
    #     def filter_by_kingdom(self, kingdom: Kingdom) -> "DarwinCoreDataFrame":
    #         return self.filter(pl.col("kingdom") == str(kingdom)).cast(DarwinCoreDataFrame)

    # Typing trick: return "Self" once Python 3.11+ only

    def __init__(self, data: pl.DataFrame | Sequence[Mapping[str, Any]] | None = None, **kwargs: Any) -> None:  # type: ignore[override]
        super().__init__(data, **kwargs) 


from pathlib import Path
import unittest
import tempfile

import polars as pl
from polars_darwin_core.lf_csv import DarwinCoreCsvLazyFrame, read_darwin_core_csv


class TestLfCsv(unittest.TestCase):
    def test_read_darwin_core_csv(self) -> None:
        # Create a tiny Darwin Core‐like CSV in a temporary directory
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)
            csv_path = tmp_path / "dwc.csv"
            csv_path.write_text("id,kingdom\n1,Animalia\n2,Plantae\n")

            lf = read_darwin_core_csv(csv_path)
            self.assertIsInstance(lf, DarwinCoreCsvLazyFrame)

            df: pl.DataFrame = lf._inner.collect()
            self.assertEqual(df.shape, (2, 2))  # two rows, two columns
            self.assertEqual(df["kingdom"].to_list(), ["Animalia", "Plantae"])



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type

import polars as pl

from polars_darwin_core.darwin_core import kingdom_data_type

__all__ = [
    "DarwinCoreCsvLazyFrame",
    "read_darwin_core_csv",
]


class DarwinCoreCsvLazyFrame:  # pylint: disable=too-few-public-methods
    """A thin wrapper around :pyclass:`polars.LazyFrame` for Darwin Core CSVs.

    The class intentionally exposes (and delegates to) the full *polars* lazy
    API while giving the object a domain-specific identity that tools like
    linters and type-checkers can understand.
    """

    # Common required fields in Darwin Core datasets
    EXPECTED_SCHEMA: Dict[str, Type[pl.DataType] | pl.DataType] = {
        # Required core fields
        "scientificName": pl.Utf8,
        "kingdom": kingdom_data_type,
        # Optional but common fields
        "phylum": pl.Utf8,
        "class": pl.Utf8,
        "order": pl.Utf8,
        "family": pl.Utf8,
        "genus": pl.Utf8,
        "species": pl.Utf8,
        # Geolocation fields
        "decimalLatitude": pl.Float64,
        "decimalLongitude": pl.Float64,
        "continent": pl.Utf8,
        "country": pl.Utf8,
        "countryCode": pl.Utf8,
        "stateProvince": pl.Utf8,
        "county": pl.Utf8,
        "municipality": pl.Utf8,
        "locality": pl.Utf8,
        "verbatimLocality": pl.Utf8,
        "minimumElevationInMeters": pl.Float64,
        "maximumElevationInMeters": pl.Float64,
        "verbatimElevation": pl.Utf8,
        "minimumDepthInMeters": pl.Float64,
        "maximumDepthInMeters": pl.Float64,
        "verbatimDepth": pl.Utf8,
        "geodeticDatum": pl.Utf8,
        "coordinateUncertaintyInMeters": pl.Float64,
        "georeferenceProtocol": pl.Utf8,
        "georeferenceSources": pl.Utf8,
        "georeferenceVerificationStatus": pl.Utf8,
        # Occurrence fields
        "basisOfRecord": pl.Utf8,
        "occurrenceID": pl.Utf8,
        "eventDate": pl.Utf8,
        "catalogNumber": pl.Utf8,
        "recordNumber": pl.Utf8,
        "recordedBy": pl.Utf8,
        "individualCount": pl.Int64,
        "sex": pl.Utf8,
        "lifeStage": pl.Utf8,
        "reproductiveCondition": pl.Utf8,
        "occurrenceStatus": pl.Utf8,
        "occurrenceRemarks": pl.Utf8,
        # Record-level fields
        "type": pl.Utf8,
        "modified": pl.Utf8,
        "language": pl.Utf8,
        "license": pl.Utf8,
        "rightsHolder": pl.Utf8,
        "accessRights": pl.Utf8,
        "bibliographicCitation": pl.Utf8,
        "references": pl.Utf8,
        "institutionID": pl.Utf8,
        "collectionID": pl.Utf8,
        "datasetID": pl.Utf8,
        "institutionCode": pl.Utf8,
        "collectionCode": pl.Utf8,
        "datasetName": pl.Utf8,
        "ownerInstitutionCode": pl.Utf8,
        "informationWithheld": pl.Utf8,
        "dataGeneralizations": pl.Utf8,
        "dynamicProperties": pl.Utf8,
    }

    # SCHEMA_OVERRIDES = {
    #     "decimalLatitude": pl.Float64(),
    #     "decimalLongitude": pl.Float64(),
    #     "taxonKey": pl.UInt64(),
    #     "verbatimScientificName": pl.String(),
    #     "order": pl.String(),
    #     "recordedBy": pl.String(),
    #     "kingdom": kingdom_enum,
    # }

    def __init__(self, inner: pl.LazyFrame):
        """Initialize the Darwin Core LazyFrame wrapper.

        Parameters
        ----------
        inner : pl.LazyFrame
            The inner LazyFrame to wrap
        """
        self._inner = inner


# -------------------------------------------------------------------------
# Convenience functions
# -------------------------------------------------------------------------


def read_darwin_core_csv(
    path: str | Path, **scan_csv_kwargs: Any
) -> DarwinCoreCsvLazyFrame:
    """Scan a Darwin Core CSV lazily.

    This is a very light wrapper around :pyfunc:`polars.scan_csv` that returns a
    domain-specific :class:`DarwinCoreCsvLazyFrame` instead of a plain
    :class:`polars.LazyFrame`.

    Parameters
    ----------
    path : str | Path
        Path to the CSV file
    **scan_csv_kwargs
        Additional keyword arguments passed to pl.scan_csv
    """

    inner = pl.scan_csv(
        path,
        schema_overrides=DarwinCoreCsvLazyFrame.EXPECTED_SCHEMA,
        quote_char=None,
        **scan_csv_kwargs,
    )
    return DarwinCoreCsvLazyFrame(inner)



import polars as pl
from enum import Enum, unique
from typing import List

__all__ = [
    "Kingdom",
    "TAXONOMIC_RANKS",
]


@unique
class Kingdom(str, Enum):
    """Enumeration of valid Darwin Core kingdoms."""

    ANIMALIA = "Animalia"
    PLANTAE = "Plantae"
    FUNGI = "Fungi"
    PROTOZOA = "Protozoa"
    CHROMISTA = "Chromista"
    ARCHAEA = "Archaea"
    BACTERIA = "Bacteria"
    VIRUSES = "Viruses"
    INCERTAE_SEDIS = "incertae sedis"

    def __str__(self) -> str:
        return str(self.value)


kingdom_data_type = pl.Enum(Kingdom)


# See: https://dwc.tdwg.org/terms/#taxon
TAXONOMIC_RANKS: List[str] = [
    "domain",
    "kingdom",
    "phylum",
    "class",
    "order",
    "family",
    "genus",
    "species",
    "subspecies",
    "variety",
    "form",
]



from pathlib import Path
import unittest

import polars as pl
from polars_darwin_core import read_darwin_core_csv, DarwinCoreCsvLazyFrame
from polars_darwin_core.archive import scan_archive
from polars_darwin_core.darwin_core import Kingdom


class TestDarwinCore(unittest.TestCase):
    def test_sample_dataset_occurrence(self) -> None:
        """Ensure we can lazily scan the sample Darwin Core occurrence file."""

        data_dir = Path(__file__).parents[1] / "sample-dataset"
        file_path = data_dir / "occurrence.txt"

        lf = read_darwin_core_csv(file_path, separator="\t")
        self.assertIsInstance(lf, DarwinCoreCsvLazyFrame)

        # Collect a small portion to verify schema and some basic stats
        # Polars will lazily evaluate; we collect first 100 rows for speed.
        df: pl.DataFrame = lf._inner.head(1).collect()
        self.assertEqual(df.shape, (1, df.width))
        expected_row = {
            "gbifID": 984586740,
            "accessRights": None,
            "bibliographicCitation": None,
            "language": None,
            "license": "CC_BY_4_0",
            "modified": None,
            "publisher": None,
            "references": None,
            "rightsHolder": None,
            "type": None,
            "institutionID": None,
            "collectionID": None,
            "datasetID": None,
            "institutionCode": "CLO",
            "collectionCode": "EBIRD",
            "datasetName": None,
            "ownerInstitutionCode": None,
            "basisOfRecord": "HUMAN_OBSERVATION",
            "informationWithheld": None,
            "dataGeneralizations": None,
            "dynamicProperties": None,
            "occurrenceID": "URN:catalog:CLO:EBIRD:OBS96797545",
            "catalogNumber": "OBS96797545",
            "recordNumber": None,
            "recordedBy": "obsr134796",
            "recordedByID": None,
            "individualCount": 1,
            "organismQuantity": None,
            "organismQuantityType": None,
            "sex": None,
            "lifeStage": None,
            "reproductiveCondition": None,
            "caste": None,
            "behavior": None,
            "vitality": None,
            "establishmentMeans": None,
            "degreeOfEstablishment": None,
            "pathway": None,
            "georeferenceVerificationStatus": None,
            "occurrenceStatus": "PRESENT",
            "preparations": None,
            "disposition": None,
            "associatedOccurrences": None,
            "associatedReferences": None,
            "associatedSequences": None,
            "associatedTaxa": None,
            "otherCatalogNumbers": None,
            "occurrenceRemarks": None,
            "organismID": None,
            "organismName": None,
            "organismScope": None,
            "associatedOrganisms": None,
            "previousIdentifications": None,
            "organismRemarks": None,
            "materialEntityID": None,
            "materialEntityRemarks": None,
            "verbatimLabel": None,
            "materialSampleID": None,
            "eventID": None,
            "parentEventID": None,
            "eventType": None,
            "fieldNumber": None,
            "eventDate": "1977-02-26",
            "eventTime": None,
            "startDayOfYear": 57,
            "endDayOfYear": 57,
            "year": 1977,
            "month": 2,
            "day": 26,
            "verbatimEventDate": None,
            "habitat": None,
            "samplingProtocol": None,
            "sampleSizeValue": None,
            "sampleSizeUnit": None,
            "samplingEffort": None,
            "fieldNotes": None,
            "eventRemarks": None,
            "locationID": None,
            "higherGeographyID": None,
            "higherGeography": None,
            "continent": "EUROPE",
            "waterBody": None,
            "islandGroup": None,
            "island": None,
            "countryCode": "MC",
            "stateProvince": "Monaco",
            "county": None,
            "municipality": None,
            "locality": "Monte-Carlo Monaco",
            "verbatimLocality": None,
            "verbatimElevation": None,
            "verticalDatum": None,
            "verbatimDepth": None,
            "minimumDistanceAboveSurfaceInMeters": None,
            "maximumDistanceAboveSurfaceInMeters": None,
            "locationAccordingTo": None,
            "locationRemarks": None,
            "decimalLatitude": 43.732407,
            "decimalLongitude": 7.422724,
            "coordinateUncertaintyInMeters": None,
            "coordinatePrecision": None,
            "pointRadiusSpatialFit": None,
            "verbatimCoordinateSystem": None,
            "verbatimSRS": None,
            "footprintWKT": None,
            "footprintSRS": None,
            "footprintSpatialFit": None,
            "georeferencedBy": None,
            "georeferencedDate": None,
            "georeferenceProtocol": None,
            "georeferenceSources": None,
            "georeferenceRemarks": None,
            "geologicalContextID": None,
            "earliestEonOrLowestEonothem": None,
            "latestEonOrHighestEonothem": None,
            "earliestEraOrLowestErathem": None,
            "latestEraOrHighestErathem": None,
            "earliestPeriodOrLowestSystem": None,
            "latestPeriodOrHighestSystem": None,
            "earliestEpochOrLowestSeries": None,
            "latestEpochOrHighestSeries": None,
            "earliestAgeOrLowestStage": None,
            "latestAgeOrHighestStage": None,
            "lowestBiostratigraphicZone": None,
            "highestBiostratigraphicZone": None,
            "lithostratigraphicTerms": None,
            "group": None,
            "formation": None,
            "member": None,
            "bed": None,
            "identificationID": None,
            "verbatimIdentification": None,
            "identificationQualifier": None,
            "typeStatus": None,
            "identifiedBy": None,
            "identifiedByID": None,
            "dateIdentified": None,
            "identificationReferences": None,
            "identificationVerificationStatus": None,
            "identificationRemarks": None,
            "taxonID": None,
            "scientificNameID": None,
            "acceptedNameUsageID": None,
            "parentNameUsageID": None,
            "originalNameUsageID": None,
            "nameAccordingToID": None,
            "namePublishedInID": None,
            "taxonConceptID": "avibase-E5531933",
            "scientificName": "Lophophanes cristatus (Linnaeus, 1758)",
            "acceptedNameUsage": None,
            "parentNameUsage": None,
            "originalNameUsage": None,
            "nameAccordingTo": None,
            "namePublishedIn": None,
            "namePublishedInYear": None,
            "higherClassification": None,
            "kingdom": Kingdom.ANIMALIA,
            "phylum": "Chordata",
            "class": "Aves",
            "order": "Passeriformes",
            "superfamily": None,
            "family": "Paridae",
            "subfamily": None,
            "tribe": None,
            "subtribe": None,
            "genus": "Lophophanes",
            "genericName": "Lophophanes",
            "subgenus": None,
            "infragenericEpithet": None,
            "specificEpithet": "cristatus",
            "infraspecificEpithet": None,
            "cultivarEpithet": None,
            "taxonRank": "SPECIES",
            "verbatimTaxonRank": None,
            "vernacularName": "Crested Tit",
            "nomenclaturalCode": None,
            "taxonomicStatus": "ACCEPTED",
            "nomenclaturalStatus": None,
            "taxonRemarks": None,
            "datasetKey": "4fa7b334-ce0d-4e88-aaae-2e0c138d049e",
            "publishingCountry": "MC",
            "lastInterpreted": "2025-02-05T00:21:14.527Z",
            "elevation": None,
            "elevationAccuracy": None,
            "depth": None,
            "depthAccuracy": None,
            "distanceFromCentroidInMeters": 245.34532311724828,
            "issue": "CONTINENT_DERIVED_FROM_COORDINATES;TAXON_MATCH_TAXON_CONCEPT_ID_IGNORED",
            "mediaType": None,
            "hasCoordinate": True,
            "hasGeospatialIssues": False,
            "taxonKey": 2487883,
            "acceptedTaxonKey": 2487883,
            "kingdomKey": 1,
            "phylumKey": 44,
            "classKey": 212,
            "orderKey": 729,
            "familyKey": 9327,
            "genusKey": 2487881,
            "subgenusKey": None,
            "speciesKey": 2487883,
            "species": "Lophophanes cristatus",
            "acceptedScientificName": "Lophophanes cristatus (Linnaeus, 1758)",
            "verbatimScientificName": "Lophophanes cristatus",
            "typifiedName": None,
            "protocol": "DWC_ARCHIVE",
            "lastParsed": "2025-02-05T00:21:14.527Z",
            "lastCrawled": "2024-09-27T13:35:39.907Z",
            "repatriated": False,
            "relativeOrganismQuantity": None,
            "projectId": None,
            "isSequenced": False,
            "gbifRegion": "EUROPE",
            "publishedByGbifRegion": "EUROPE",
            "level0Gid": "MCO",
            "level0Name": "Monaco",
            "level1Gid": None,
            "level1Name": None,
            "level2Gid": None,
            "level2Name": None,
            "level3Gid": None,
            "level3Name": None,
            "iucnRedListCategory": "LC",
        }

        self.assertEqual(df.to_dicts()[0], expected_row)

    def test_scan_archive_core(self) -> None:
        """Test the scan_archive function using the existing sample dataset."""
        # Use the sample dataset directory from the repository
        data_dir = Path(__file__).parents[1] / "sample-dataset"

        # Call scan_archive with the sample dataset
        lf = scan_archive(data_dir)

        # Validate the result is a DarwinCoreCsvLazyFrame
        self.assertIsInstance(lf, DarwinCoreCsvLazyFrame)

        # Collect a small portion to verify we can access the data
        df = lf._inner.head(2).collect()

        # Verify we have the right schema - occurrence.txt has many columns
        self.assertTrue(df.width > 10)

        # Check for some expected columns from Darwin Core
        expected_columns = [
            "scientificName",
            "kingdom",
            "decimalLatitude",
            "decimalLongitude",
        ]
        for col in expected_columns:
            self.assertIn(col, df.columns)

        # Verify we can access a specific value from the first row
        # The first record should be for "Lophophanes cristatus" based on our earlier test
        self.assertEqual(
            df["scientificName"][0], "Lophophanes cristatus (Linnaeus, 1758)"
        )



__version__ = "1.0.0"

"""Top-level package for polars_darwin_core.

This library provides helpers for working with Darwin Core (DwC) data in
polars DataFrames and LazyFrames.
"""

from .darwin_core import Kingdom, kingdom_data_type, TAXONOMIC_RANKS
from .lf_csv import (
    DarwinCoreCsvLazyFrame,
    read_darwin_core_csv,
)
from .archive import scan_archive

__all__ = [
    "__version__",
    "Kingdom",
    "kingdom_data_type",
    "TAXONOMIC_RANKS",
    "DarwinCoreCsvLazyFrame",
    "read_darwin_core_csv",
    "scan_archive",
]
